module Json.Decode.BrokenSpec exposing
    ( decodeToRecord
    , parseArray
    , parseFalse
    , parseNothing
    , parseNull
    , parseNumber
    , parseObject
    , parseString
    , parseTrue
    )

import Expect
import Fuzz
import Json.Decode as Decode
import Json.Decode.Broken as Broken
import Json.Encode as Encode
import Parser
import Test exposing (..)


parseAndDecode : String -> Decode.Decoder a -> Result String a
parseAndDecode json decoder =
    case Broken.parse json of
        Ok value ->
            Decode.decodeValue decoder value
                |> Result.mapError Decode.errorToString

        Err error ->
            Err <| Debug.toString error


parseDecodeAndEncode : String -> Result String String
parseDecodeAndEncode json =
    parseAndDecode json Decode.value
        |> Result.map (Encode.encode 0)


parseNothing : Test
parseNothing =
    test "parse the empty string" <|
        -- Ignore the details of the error message.
        \_ -> Broken.parse "" |> Result.mapError (always Nothing) |> Expect.equal (Err Nothing)


parseNull : Test
parseNull =
    test "parse literal 'null'" <|
        \_ -> parseAndDecode "null" (Decode.null ()) |> Expect.equal (Ok ())


parseFalse : Test
parseFalse =
    test "parse literal 'false'" <|
        \_ -> parseAndDecode "false" Decode.bool |> Expect.equal (Ok False)


parseTrue : Test
parseTrue =
    test "parse literal 'true'" <|
        \_ -> parseAndDecode "true" Decode.bool |> Expect.equal (Ok True)


parseNumber : Test
parseNumber =
    describe "parse numbers"
        [ fuzz Fuzz.float "integers" <|
            \f ->
                parseAndDecode (String.fromFloat f) Decode.float |> Expect.equal (Ok f)
        , test "integer with exponent" <|
            \_ ->
                parseAndDecode "123e45" Decode.float |> Expect.equal (Ok 1.23e47)
        , test "integer with plus-signed exponent" <|
            \_ ->
                parseAndDecode "123e+45" Decode.float |> Expect.equal (Ok 1.23e47)
        , test "integer with minus-signed exponent" <|
            \_ ->
                parseAndDecode "123e-45" Decode.float |> Expect.equal (Ok 1.23e-43)
        , test "float" <|
            \_ ->
                parseAndDecode "123.456" Decode.float |> Expect.equal (Ok 123.456)
        , test "float with exponent" <|
            \_ ->
                parseAndDecode "123.456e78" Decode.float |> Expect.equal (Ok 1.23456e80)
        , test "float with plus-signed exponent" <|
            \_ ->
                parseAndDecode "123.456e+78" Decode.float |> Expect.equal (Ok 1.23456e80)
        , test "float with minus-signed exponent" <|
            \_ ->
                parseAndDecode "123.456e-78" Decode.float |> Expect.equal (Ok 1.23456e-76)
        ]


parseString : Test
parseString =
    describe "parse strings"
        [ test "simple" <|
            \_ ->
                parseAndDecode "\"simple\"" Decode.string |> Expect.equal (Ok "simple")
        , test "escape quote" <|
            \_ ->
                parseAndDecode "\"\\\"\"" Decode.string |> Expect.equal (Ok "\"")
        , test "escape reverse solidus (backslash)" <|
            \_ ->
                parseAndDecode "\"\\\\\"" Decode.string |> Expect.equal (Ok "\\")
        , test "escape solidus (forward-slash)" <|
            \_ ->
                parseAndDecode "\"\\/\"" Decode.string |> Expect.equal (Ok "/")
        , test "escape backspace" <|
            \_ ->
                parseAndDecode "\"\\b\"" Decode.string |> Expect.equal (Ok "\u{0008}")
        , test "escape formfeed" <|
            \_ ->
                parseAndDecode "\"\\f\"" Decode.string |> Expect.equal (Ok "\u{000C}")
        , test "escape newline" <|
            \_ ->
                parseAndDecode "\"\\n\"" Decode.string |> Expect.equal (Ok "\n")
        , test "escape carriage return" <|
            \_ ->
                parseAndDecode "\"\\r\"" Decode.string |> Expect.equal (Ok "\u{000D}")
        , test "escape horizontal tab" <|
            \_ ->
                parseAndDecode "\"\\t\"" Decode.string |> Expect.equal (Ok "\t")
        , test "escape unicode" <|
            \_ ->
                parseAndDecode "\"\\u0040\"" Decode.string |> Expect.equal (Ok "@")
        , test "escape mixed" <|
            \_ ->
                parseAndDecode "\" \\\\ \\/ \\b \\f \\n \\r \\t \\u0040 \"" Decode.string
                    |> Expect.equal (Ok " \\ / \u{0008} \u{000C} \n \u{000D} \t @ ")
        ]


parseArray : Test
parseArray =
    describe "parse arrays"
        [ test "empty" <|
            \_ ->
                parseAndDecode "[]" (Decode.list Decode.value) |> Expect.equal (Ok [])
        , test "with single element" <|
            \_ ->
                parseAndDecode "[null]" (Decode.list <| Decode.null ()) |> Expect.equal (Ok [ () ])
        , test "with multiple elements" <|
            \_ ->
                parseAndDecode "[null, true, false, 1234, \"foo\"]" (Decode.list Decode.value)
                    |> Result.map (List.map (Encode.encode 2))
                    |> Expect.equal
                        (Ok
                            [ "null"
                            , "true"
                            , "false"
                            , "1234"
                            , "\"foo\""
                            ]
                        )
        , test "with nested values" <|
            \_ ->
                parseDecodeAndEncode "[ [null, true], [false, 1234], {\"foo\": \"bar\"} ]"
                    |> Expect.equal (Ok "[[null,true],[false,1234],{\"foo\":\"bar\"}]")
        ]


parseObject : Test
parseObject =
    describe "parse objects"
        [ test "empty" <|
            \_ ->
                parseDecodeAndEncode "{   }"
                    |> Expect.equal (Ok "{}")
        , test "with single member (object)" <|
            \_ ->
                parseDecodeAndEncode "{ \"foo\" : {} }"
                    |> Expect.equal (Ok "{\"foo\":{}}")
        , test "with single member (array)" <|
            \_ ->
                parseDecodeAndEncode "{ \"foo\" : [] }"
                    |> Expect.equal (Ok "{\"foo\":[]}")
        , test "with single member (string)" <|
            \_ ->
                parseDecodeAndEncode "{\"foo\" :\"bar\"}"
                    |> Expect.equal (Ok "{\"foo\":\"bar\"}")
        , test "with single member (number)" <|
            \_ ->
                parseDecodeAndEncode "{\"foo\" \u{000D}: \t 123\n}"
                    |> Expect.equal (Ok "{\"foo\":123}")
        , test "with single member (true)" <|
            \_ ->
                parseDecodeAndEncode "{\"foo\": true}"
                    |> Expect.equal (Ok "{\"foo\":true}")
        , test "with single member (false)" <|
            \_ ->
                parseDecodeAndEncode "{\"foo\": false}"
                    |> Expect.equal (Ok "{\"foo\":false}")
        , test "with single member (null)" <|
            \_ ->
                parseDecodeAndEncode "{\"foo\": null}"
                    |> Expect.equal (Ok "{\"foo\":null}")
        ]


type alias Record =
    { aaa : List String
    , bbb : Float
    , ccc : Bool
    , ddd : Bool
    , eee : ()
    }


decodeToRecord : Test
decodeToRecord =
    let
        json =
            "{\"aaa\" : [ \"foo\", \"bar\" ] , \"bbb\" : 12.3e4 , "
                ++ "\"ccc\": true , \"ddd\"   :false , \"eee\" : null }"

        decoder =
            Decode.map5 Record
                (Decode.field "aaa" (Decode.list Decode.string))
                (Decode.field "bbb" Decode.float)
                (Decode.field "ccc" Decode.bool)
                (Decode.field "ddd" Decode.bool)
                (Decode.field "eee" (Decode.null ()))
    in
    test "decode to an Elm record" <|
        \_ ->
            parseAndDecode json decoder
                |> Expect.equal
                    (Ok
                        { aaa = [ "foo", "bar" ]
                        , bbb = 123000
                        , ccc = True
                        , ddd = False
                        , eee = ()
                        }
                    )



-- Local Variables:
-- elm-format-elm-version: "0.19"
-- End:
